/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail;


//import java
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

//import javax
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.MBeanServer;
import javax.management.MBeanRegistrationException;
import javax.management.NotCompliantMBeanException;

//import objectweb
import org.objectweb.jonas.common.JProp;
import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jmx.JmxService;
import org.objectweb.jonas.jmx.JonasObjectName;
import org.objectweb.jonas.service.AbsServiceImpl;
import org.objectweb.jonas.service.ServiceException;
import org.objectweb.jonas.service.ServiceManager;

//import objectweb.util
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

//import jonas
import org.objectweb.jonas.mail.factory.JMailSession;
import org.objectweb.jonas.mail.factory.JMailMimePartDS;


/**
 * This class provides an implementation of the javaMail service.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class MailServiceImpl extends AbsServiceImpl
    implements MailService, MailServiceImplMBean {

    /**
     * The list of the registered mail session factories.
     */
    private Vector mailSessionList = null;

    /**
     * The list of the registered mime part data source factories.
     */
    private Vector mailMimePartDSList = null;

    /**
     * Reference to a MBean server.
     */
    private MBeanServer mbeanServer = null;

   /**
    * Initial Context for Naming
    */
    private Context ictx = null;

    /**
     * Logger for this service.
     */
    private static Logger logger = null;

    /**
     * List of the factories names to load when starting the MailService
     */
    private Vector factoryNames = new Vector();

    /**
     *  List of the factories bound from the javax.mail.Session type
     */
    private Hashtable jMailSessionFactories = new Hashtable();

    /**
     *  List of the factories bound from the javax.mail.internet.MimePartDataSource type
     */
    private Hashtable jMailMimePartDSFactories = new Hashtable();


    /**
     * Init the Mail service.
     * @param ctx the configuration of the Ear service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doInit(Context ctx) throws ServiceException {

        // get logger for this service
        logger = Log.getLogger(Log.JONAS_MAIL_PREFIX);

        //Get the inital Context
        try {
            ictx = new InitialContext();
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR, "Cannot create initial context during the mail service initializing");
            throw new ServiceException("Cannot create initial context during the mail service initializing", e);
        }

        // Get the JMX Server via JMX Service
        try {
            mbeanServer = 
                ((JmxService) ServiceManager.getJmxService()).getJmxServer();
        } catch (ServiceException e) {
            // the JMX service may not be started
            mbeanServer = null;
        }
        
        // Get the list of the factory names
        String factories = null;
        try {
            factories = (String) ctx.lookup("factories");
        } catch (NamingException e) {
            // No problem if there is no value for 'factories'
        }
        if (factories != null) {
            StringTokenizer st = new StringTokenizer(factories, ",");
            while (st.hasMoreTokens()) {
                factoryNames.add(st.nextToken().trim());
            }
        }
        logger.log(BasicLevel.DEBUG, "Mail service initialized");
     
    }

    /**
     * Start the Mail Service.
     * @throws ServiceException if the initialization failed.
     */
    protected void doStart() throws ServiceException {

        // creates each factory
        String factoryName = null;
        for (int i = 0; i < factoryNames.size(); i++) {
            factoryName = (String) factoryNames.elementAt(i);
            try {
                JProp prop = new JProp(factoryName);
                logger.log(BasicLevel.DEBUG, "Creating mail factory " + factoryName);
                createMailFactory(prop.getFilesEnv());
            } catch (Exception e) {
                logger.log(BasicLevel.ERROR, "JOnAS: Cannot create mail factory : " + factoryName + " " + e);
                logger.log(BasicLevel.ERROR, "Please check if " + factoryName + ".properties is available");
            }
        }

        try {
            // Register MailService MBean : MailServiceImplMBean
            if (mbeanServer != null) {
                mbeanServer.registerMBean(this, JonasObjectName.mailService());
            }
        } catch (InstanceAlreadyExistsException iae) {
            throw new ServiceException("Cannot start the MailService (Already Exists)" , iae);
        } catch (MBeanRegistrationException mbre) {
            throw new ServiceException("Cannot start the MailService (MBean registration error)" , mbre);
        } catch (NotCompliantMBeanException ncmbe) {
            throw new ServiceException("Cannot start the MailService (MBean Not compliant error)" , ncmbe);
        }


    }
    
    /**
     * Stop the Mail service.
     * @throws ServiceException if the stop failed.
     */
    protected void doStop() throws ServiceException {

        try {
            unbindMailFactories();
        } catch (MailServiceException e) {
            logger.log(BasicLevel.ERROR, "Cannot unbind mail factories " + e);
            throw new ServiceException("Cannot unbind mail factories ", e);
        }


    }

    /**
     * Create a mail factory with the specified properties and register it 
     * into the registry.
     * @param props the properties used to configure the mail factory.
     * @throws MailServiceException if the creation or the registration of 
     * the factory failed.
     */
    public void createMailFactory(Properties props) 
            throws MailServiceException {

        Object factory = null;

        //Factory type/name
        String mailFactoryName = props.getProperty("mail.factory.name");
        String mailFactoryType = props.getProperty("mail.factory.type");

        //Define the name of the factory
        if (mailFactoryName == null) {
            throw new MailServiceException("The property 'mail.factory.name' is a required property for this factory. It must be set in the properties file");
        }

        if (mailFactoryType == null) {
            throw new MailServiceException("The property 'mail.factory.type' is a required property for this factory. It must be set in the properties file");
        }

        if ((jMailSessionFactories.get(mailFactoryName) != null) || (jMailMimePartDSFactories.get(mailFactoryName) != null)) {
            throw new MailServiceException("There is already a factory bound with the name '" + mailFactoryName +  "', please correct your properties.file");
        }

        //Two kind of factory
        final int JAVAX_MAIL_SESSION_FACTORY = 1;
        final int JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE = 2;

        //Define the type of our factory
        int typeOfFactory;
        if (mailFactoryType.equalsIgnoreCase("javax.mail.Session")) {
            typeOfFactory = JAVAX_MAIL_SESSION_FACTORY;
        } else if (mailFactoryType.equalsIgnoreCase("javax.mail.internet.MimePartDataSource")) {
            typeOfFactory = JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE;
        } else {
            typeOfFactory = 0;
        }
        
        switch (typeOfFactory) {
            
        case JAVAX_MAIL_SESSION_FACTORY :
            JMailSession jMailSession = new JMailSession(mailFactoryName, props);
            //add the factory to the list
            jMailSessionFactories.put(mailFactoryName, jMailSession);
            factory = jMailSession;
            break;
            
        case JAVAX_MAIL_INTERNET_MIMEPARTDATASOURCE :
            JMailMimePartDS jMailMimePartDs = new JMailMimePartDS(mailFactoryName, props);

            //Add to the vector list
            jMailMimePartDSFactories.put(mailFactoryName, jMailMimePartDs);
            factory = jMailMimePartDs;
            break;
        default :
            throw new MailServiceException("Can not create a factory of the type '" + mailFactoryType + "'. This type is incorrect");
        }

        try {
            //The factory is in the factory object.
            //Name for the bind is in mailFactoryName var.
            ictx.rebind(mailFactoryName, factory);
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR, "Cannot bind mail factory '" + mailFactoryName + "'  :" + e.getMessage());
            throw new MailServiceException("Cannot unbind mail factories ", e);
        }
        logger.log(BasicLevel.WARN, "Mapping Mail Factory "+mailFactoryType+"  on "+ mailFactoryName);
    }

    /**
     * Create a mail factory with the specified properties and register it 
     * into the registry.
     * @param props the properties used to configure the mail factory.
     * @throws MailServiceException if the creation or the registration of 
     * the factory failed.     
     * @throws RemoteException if rmi call failed.
     */
    public void createMailFactoryMBean(Properties props) 
            throws RemoteException, MailServiceException {

        //call the internal proc
        createMailFactory(props);
    }

    /**
     * Unregister all the binding factories on the server.
     * @throws MailServiceException if the unregistration of the factories 
     * failed.
     */
    public void unbindMailFactories() throws MailServiceException {

        String nameOfFactory = null;
        //Unbind all factories
        try {
            //First unbind the Mail session factories
            for (Enumeration e = jMailSessionFactories.elements(); e.hasMoreElements();) {
                JMailSession jMailSession = (JMailSession) e.nextElement();
                nameOfFactory = jMailSession.getName();
                ictx.unbind(nameOfFactory);
            }

            //And now unbind the mime part datasource factories
            for (Enumeration e = jMailMimePartDSFactories.elements(); e.hasMoreElements();) {
                JMailMimePartDS jMailMimePartDS  = (JMailMimePartDS) e.nextElement();
                nameOfFactory = jMailMimePartDS.getName();
                ictx.unbind(nameOfFactory);
            }
        } catch (NamingException e) {
            logger.log(BasicLevel.ERROR, "cannot unbind factories : " + e.getMessage());
            throw new MailServiceException("Can not unbind the factories.", e);
        }

        //Clear the hashtable
        jMailSessionFactories.clear();
        jMailMimePartDSFactories.clear();
        
    }

    /**
     * Unregister the binding factory with the given name.
     * @param name the name of the factory.
     * @throws RemoteException if rmi call failed.
     * @throws MailServiceException if the unregistration of the factory 
     * failed.
     */
    public void unbindMailFactoryMBean(String name) 
            throws RemoteException, MailServiceException {

        JMailSession jmailSession = (JMailSession) jMailSessionFactories.get(name);
        JMailMimePartDS jMailMimePartDS = (JMailMimePartDS) jMailMimePartDSFactories.get(name);

        if (jmailSession != null) {
            System.out.println("Mail Factory of type javax.mail.session : " + name + " removed");
            jMailSessionFactories.remove(name);
        } else if (jMailMimePartDS != null) {
            System.out.println("Mail Factory of type javax.mail.internet.MimePartDataSource : " + name + " removed");
            jMailMimePartDSFactories.remove(name);
        } else {
            throw new MailServiceException("Can not unbind the factories. The name '" + name + "' is not the name of a bound factory");
        }

        try {
            ictx.unbind(name);
        } catch (NamingException e) {
            throw new MailServiceException("Can not unbind the factory '" + name + "'.", e);
        }
    }

    /**
     * Get the list of the registered JMailSessionFactories.
     * @return the list of the registered JMailSessionFactories.
     * @throws RemoteException if rmi call failed.
     */
    public JMailSession[] getJMailSessionFactoriesMBean() throws RemoteException {

        //Return the values of the hashtable as an array
        JMailSession[] jMailSessionArray = new JMailSession[jMailSessionFactories.size()];
        int i = 0;
        for (Enumeration e = jMailSessionFactories.elements(); e.hasMoreElements(); i++) {
            JMailSession jMailSession = (JMailSession) e.nextElement();
            jMailSessionArray[i] = jMailSession;
        }
        return jMailSessionArray;
    }

    /**
     * Get the list of the registered JMailMimePartDSFactories.
     * @return the list of the registered JMailMimePartDSFactories.
     * @throws RemoteException if rmi call failed.
     */
    public JMailMimePartDS[] getJMailMimePartDSFactoriesMBean() throws RemoteException {

        //Return the values of the hashtable as an array
        JMailMimePartDS[] jMailMimePartDSArray = new JMailMimePartDS[jMailMimePartDSFactories.size()];
        int i = 0;
        for (Enumeration e = jMailMimePartDSFactories.elements(); e.hasMoreElements(); i++) {
            JMailMimePartDS jMailMimePartDS = (JMailMimePartDS) e.nextElement();
            jMailMimePartDSArray[i] = jMailMimePartDS;
        }
        return jMailMimePartDSArray;
    }
    
}
