/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or 1any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail.factory;

//import java
import java.io.ObjectOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Properties;
import org.apache.regexp.RE;
import org.apache.regexp.RESyntaxException;

//import javax
import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;


/**
 * This class gets the properties from mail factory properties file
 * and build a properties object for the Session.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public abstract class JMail implements Serializable, Referenceable {

    /**
     * Properties of the mail factory properties file
     */
    protected Properties mailProperties = null;

    /**
     * Properties for the javax.mail.Session object.
     * Keep only the var used for configuring the javax.mail.Session object.
     */
    protected Properties mailSessionProperties = null;
    
    /**
     * Properties for the authentication.
     */
    protected Properties authenticationProperties = null;

    /**
     * List of all matching properties for configure the javax.mail.Session object
     */
    protected static final String[] PATTERN_MATCHING_SESSION = 
        new String[] { "^mail.debug$",
                       "^mail.from$",
                       "^mail.mime.address.strict$",
                       "^mail.host$",
                       "^mail.store.protocol$",
                       "^mail.transport.protocol$",
                       "^mail.user$",
                       "^mail.(.*).class$",
                       "^mail.(.*).host$",
                       "^mail.(.*).port$",
                       "^mail.(.*).user$"
        };
    
    /**
     * List of all matching properties for configure the
     * authentication of the javax.mail.Session objects.
     */
    protected static final String[] PATTERN_MATCHING_AUTHENTICATION = 
        new String[] { "^mail.authentication.username$",
                       "^mail.authentication.password$"
        };

    /**
     * The name of this object.
     */
    private String name = null;

    /**
     * Constructor of a JMail Object with the given name and properties.
     * @param name the name of this object.
     * @param mailProperties properties for configuring this object.
     */
    public JMail(String name, Properties mailProperties) {
        this.name = name;
        this.mailProperties = mailProperties;
        this.mailSessionProperties = findMailSessionProperties(mailProperties);
        this.authenticationProperties = findAuthenticationProperties(mailProperties);
    }

    /**
     * Return the name of this object
     * @return name of this Jmail.
     */
    public String getName() {
        return name;
    }

    /**
     * Return all the properties required for the configuration of
     * javax.mail.Session objects.
     * Used a pattern matching for find the good properties.
     * @param props all the properties in which, javax.mail.Session properties
     * must be found.
     * @return all the properties required for the configuration of Session objects.
     */
    protected Properties findMailSessionProperties(Properties props) {
        Properties sessionProps = new Properties();
        String key = null;

        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            //Check if the property is matching
            key = (String) e.nextElement();
            if (isMatchingSessionProperty(key)) {
                sessionProps.put(key, props.getProperty(key));
            }
        }
        return sessionProps;
    }
    
    /**
     * Return all the properties required for the configuration of
     * the authentication for the javax.mail.Session objects.
     * Used a pattern matching for find the good properties.
     * @param props all the properties defined in the properties file.
     * @return all the properties required for the configuration of Session objects.
     */
    protected Properties findAuthenticationProperties(Properties props) {
        Properties authProps = new Properties();
        String key = null;

        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            //Check if the property is matching
            key = (String) e.nextElement();
            if (isMatchingAuthenticationProperty(key)) {
                authProps.put(key, props.getProperty(key));
            }
        }
        return authProps;
    }

     /**
     * Return if the property is a property of the specified array.
     * @param property to check.
     * @param regexps in order to do the check.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingProperty(String property, String[] regexps) {
        boolean matching = false;
        int i = 0;
        RE r = null;

        //In JDK 1.4, we could use string.matches() method

        while (i < regexps.length && !matching) {
            //compare for each regexp
            try {
                r = new RE(regexps[i]);
            } catch (RESyntaxException res) {
                return false;
            }
            if (r.match(property)) {
                matching = true;
            }
            i++;
        }
        return matching;
    }

    /**
     * Return if the property is a javax.mail.Session property.
     * @param property to do the check with this property.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingSessionProperty(String property) {
        return isMatchingProperty(property, PATTERN_MATCHING_SESSION);
    }

    /**
     * Return if the property is a authentication property.
     * @param property to do the check with this property.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingAuthenticationProperty(String property) {
        return isMatchingProperty(property, PATTERN_MATCHING_AUTHENTICATION);
    }

    /**
     * Retrieves the Reference of the object. 
     * The Reference contains the factory used to create this object 
     * and the optional parameters used to configure the factory.
     * @return the non-null Reference of the object. 
     * @throws NamingException if a naming exception was encountered while 
     * retrieving the reference.
     */
    public abstract Reference getReference() throws NamingException;


    /**
     * Retrieves all the properties of this object. 
     * @return the properties of this object.
     */
    public Properties getProperties() {
        return mailProperties;
    }

    /**
     * Retrieves the session properties of this object. 
     * @return the session properties of this object.
     */
    public Properties getMailSessionProperties() {
        return mailSessionProperties;
    }

    /**
     * Retrieves the authentication properties of this object. 
     * @return the authentication properties of this object.
     */
    public Properties getAuthenticationProperties() {
        return authenticationProperties;
    }


    /**
     * Return an array of byte from a given object
     * @param obj the object from which we must extract the bytes.
     * @return the byte[] from an object
     */
    protected byte[] getBytesFromObject(Object obj) {

        if (obj == null) {
            return null;
        }

        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        
        ObjectOutputStream oos = null;
        byte[] bytes = null;
      
        try {
            oos = new ObjectOutputStream(baos);
            oos.writeObject(obj);
            bytes = baos.toByteArray();
        } catch (Exception e) {
            //Can't tranform
            return null;
        } finally {
            try {
                oos.close();
                baos.close();
            } catch (Exception e) {
                //Not a problem
            }
        }
        return bytes;
    }
}

