/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas.mail.factory;

//import java
import java.util.Enumeration;
import java.util.Properties;

//import javax
import javax.naming.BinaryRefAddr;
import javax.naming.NamingException;
import javax.naming.Reference;

/**
 * This class provides a way for referencing mail session.
 * @author Florent Benoit
 * @author Ludovic Bert
 */
public class JMailMimePartDS extends JMail {

    /**
     * Properties for the javax.mail.MimeMessage object.
     */
    private Properties messageProperties = null;

   /**
     * List of all matching properties for configure the javax.mail.internet.MimeMessage object
     */
    protected static final String[] PATTERN_MATCHING_MIMEMESSAGE =
        new String[] { 
            "^mail.to$",
            "^mail.cc$",
            "^mail.bcc$",
            "^mail.subject$"
        };
 

    /**
     * Constructor of a JMimePartDS Object
     * @param name the name of this object.
     * @param mailProperties properties for configuring this object.
     */
    public JMailMimePartDS(String name, Properties mailProperties) {
        super(name, mailProperties);
        //Get the message properties
        messageProperties = findMimeMessageProperties(mailProperties);
    }


    /**
     * Retrieves the mimeMessage properties of this object. 
     * @return the mimeMessage properties of this object.
     */
    public Properties getMimeMessageProperties() {
        return messageProperties;
    }

    /**
     * Retrieves the Reference of the javax.mail.MimePartDataSource object. 
     * The Reference contains the factory used to create this object 
     * (that is the JMimePartDSFactory) and the optional parameters used to 
     * configure the factory.
     * @return the non-null Reference of the javax.mail.MimePartDataSource 
     * object. 
     * @throws NamingException if a naming exception was encountered while 
     * retrieving the reference.
     */
    public Reference getReference() throws NamingException {

        //Build the reference to the factory JMailSessionFactory
        Reference reference = new Reference("javax.mail.internet.MimePartDataSource",
                                            "org.objectweb.jonas.mail.factory.JMailMimePartDSFactory", 
                                            null);
        byte[] bytes = null;

        //Give the session properties
        bytes = getBytesFromObject(mailSessionProperties);
        if (bytes != null) {
            reference.add(new BinaryRefAddr("javaxmailSession.properties", bytes));
        }

        //Give the mime message properties
        bytes = getBytesFromObject(messageProperties);
        if (bytes != null) {
            reference.add(new BinaryRefAddr("javaxInternetMimeMessage.properties", bytes));
        }

        //Give the authentication properties
        bytes = getBytesFromObject(authenticationProperties);
        if (bytes != null) {
            reference.add(new BinaryRefAddr("authentication.properties", bytes));
        }

        return reference;
    }


    /**
     * Return all the properties required for the configuration of
     * javax.mail.internet.MimeMessage objects.
     * Used a pattern matching for find the good properties.
     * @param props all the properties in which, javax.mail.internet.MimeMessage properties
     * must be found.
     * @return properties used by this JMailMimePartDS
     */
    protected Properties findMimeMessageProperties(Properties props) {
        Properties mimemessageProps = new Properties();
        String key = null;

        for (Enumeration e = props.propertyNames(); e.hasMoreElements();) {
            //Check if the property is matching
            key = (String) e.nextElement();
            if (isMatchingMimeMessageProperty(key)) {
                mimemessageProps.put(key, props.getProperty(key));
            }
        }
        return mimemessageProps;
    }



    /**
     * Return if the property is a javax.mail.internet.MimeMessage property.
     * @param property to check.
     * @return true if the property is matching, else false.
     */
    protected boolean isMatchingMimeMessageProperty(String property) {
        return isMatchingProperty(property, PATTERN_MATCHING_MIMEMESSAGE);
    }


}
