/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Florent BENOIT & Ludovic BERT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.server;


//import java
import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.util.StringTokenizer;

//import jonas
import org.objectweb.jonas.common.JProp;


/**
 * This class load all the jars needed to start JOnAS and after this, it 
 * launch the JOnAS server (Server class).
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class Bootstrap {

    /**
     * The name of this server.
     */
    private static final String SERVER_NAME = "JOnAS";

    /**
     * The CLASSPATH property.
     */
    private static final String CLASSPATH = "java.class.path";

    /**
     * The JONAS_ROOT property.
     */
    private static final String JONAS_ROOT = "install.root";

    /**
     * The OBJECTWEB_ORB property.
     */
    private static final String OBJECTWEB_ORB = "objectweb.orb";

    /**
     * The JONAS_ROOT directory.
     */
    private static String jonasRoot = null;

    /**
     * The TOMCAT_HOME property.
     */
    private static final String TOMCAT_HOME = "tomcat.home";

    /**
     * The CATALINA_HOME property.
     */
    private static final String CATALINA_HOME = "catalina.home";

    /**
     * The JETTY_HOME property.
     */
    private static final String JETTY_HOME = "jetty.home";

    /**
     * A memory representation of the configuration file of the JOnAS server.
     */
    private static JProp props = null;

    /**
     * The jonas.services property key.
     */
    private static final String SERVICES_PROP_NAME = JProp.JONASPREFIX + ".services";
    
    /**
     * The list of the services used by the JOnAS server.
     */
    private static String[] services = null;
    

    /**
     * Get if the specified service is launch by the user.
     * @param serviceName the service name to get if is launched by the user.
     * @return true if only if the specified service is launch by the user.
     * @throws Exception if an error occurs when getting the services name.
     */
    private static boolean isLaunchService(String serviceName) 
        throws Exception {
        
        if (props == null) {
            props = new JProp();
            services = props.getValueAsArray(SERVICES_PROP_NAME);
        }
        for (int i = 0; i < services.length; i++) {
            if (services[i].equals(serviceName)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Return an array containing all the resources declared in the classpath
     * environment variable.
     * @return a list containing all the resources declared in the classpath
     * environment variable.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getClassPath() throws Exception {
        String cpValue = System.getProperty(CLASSPATH);
        String sep = File.pathSeparator;
        JURLs jurls = new JURLs();
        StringTokenizer tokenizer = new StringTokenizer(cpValue, sep);
        while (tokenizer.hasMoreTokens()) {
            File file = new File(tokenizer.nextToken());
            if (file.isDirectory()) {
                jurls.addDir(file);
            } else {
                jurls.add(file);
            }
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used by the JOnAS server.
     * @return a list containing all the resources used by the JOnAS server.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJonasResources() throws Exception {
        // Get the jonas.root directory.
        String jonasRoot = System.getProperty(JONAS_ROOT);
        if (jonasRoot == null) {
            throw new Exception(JONAS_ROOT + " is not set");
        }

        // Get the jonas/lib directory.
        File jonasLib = new File(jonasRoot + File.separator + "lib");
        
        // add mail.jar and activation.jar the order is important.
        JURLs jurls = new JURLs();
        if (isLaunchService("mail")) {
            jurls.add(new File(jonasLib + File.separator + "j2ee", "mail.jar"));
            jurls.add(new File(jonasLib + File.separator + "j2ee", "activation.jar"));
        }

        String objectwebOrb = System.getProperty(OBJECTWEB_ORB);
        if (objectwebOrb == null) {
            throw new Exception(OBJECTWEB_ORB + " is not set");
        }        
        jurls.add(new File(jonasLib, objectwebOrb + "_jonas.jar"));
        jurls.addNotEndWith(jonasLib, "_jonas.jar");
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Tomcat.
     * @param tomcatHomeDir the home directory of Tomcat.
     * @return a list containing all the resources used for the integration
     * of Tomcat.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getTomcatResources(File tomcatHomeDir) 
        throws Exception {
        
        JURLs jurls = new JURLs();
        String tomcatContainerDir = tomcatHomeDir + File.separator + "lib" + File.separator + "container";
        String[] tomcatContainerDirs = { 
            tomcatContainerDir + File.separator + "tomcat_util.jar",
            tomcatContainerDir + File.separator + "tomcat_modules.jar",
            tomcatContainerDir + File.separator + "facade22.jar",
            tomcatContainerDir + File.separator + "jasper.jar",
        };
        for (int i = 0; i < tomcatContainerDirs.length; i++) {
            jurls.add(new File(tomcatContainerDirs[i]));
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Catalina.
     * @param catalinaHomeDir the home directory of Catalina.
     * @return a list containing all the resources used for the integration
     * of Catalina.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getCatalinaResources(File catalinaHomeDir) 
        throws Exception {
                    
        JURLs jurls = new JURLs();
        String[] catalinaLibDirs = { "bin" + File.separator + "bootstrap.jar", 
                                     "server" + File.separator + "lib",
                                     "common" + File.separator + "lib",
                                     "lib" };
        for (int i = 0; i < catalinaLibDirs.length; i++) {
            jurls.add(new File(catalinaHomeDir, catalinaLibDirs[i]));
        }
        return jurls;
    }

    /**
     * Return an array containing all the resources used for the integration
     * of Jetty.
     * @param jettyHomeDir the home directory of Jetty.
     * @return a list containing all the resources used for the integration
     * of Jetty.
     * @throws Exception if a resource cannot be found.
     */
    private static JURLs getJettyResources(File jettyHomeDir) 
        throws Exception {

        JURLs jurls = new JURLs();
        String jettyLibDir = jettyHomeDir + File.separator + "lib";
        String[] jettyLibDirs = { 
            jettyLibDir + File.separator + "org.mortbay.jetty.jar",
            jettyLibDir + File.separator + "org.apache.jasper.jar" 
        };
        for (int i = 0; i < jettyLibDirs.length; i++) {
            jurls.add(new File(jettyLibDirs[i]));
        }        
        return jurls;
    }    

    /**
     * Get the list of the resources to load when launching JOnAS.
     * @return the list of the resources to load when launching JOnAS.
     * @throws Exception if a resource cannot be found.
     */
    public static URL[] getResources() throws Exception {
        JURLs jurls = new JURLs();

        // add resources from the classpath.
        jurls.merge(getClassPath());


        // add resources of JOnAS.
        jurls.merge(getJonasResources());

        // add resources for "jetty" service if required.
        if (isLaunchService("jetty")) {
            String jettyHome = System.getProperty(JETTY_HOME);
            if (jettyHome != null) {
                File jettyHomeDir = new File(jettyHome);
                jurls.merge(getJettyResources(jettyHomeDir));
            }
        }

        // add resources for "tomcat" service if required.
        if (isLaunchService("tomcat")) {
            String s = props.getValue("jonas.service.tomcat.class");
            if (s.endsWith("33")) {
                // the user want to use tomcat3.3
                String tomcatHome = System.getProperty(TOMCAT_HOME);
                if (tomcatHome != null) {
                    File tomcatHomeDir = new File(tomcatHome);
                    jurls.merge(getTomcatResources(tomcatHomeDir));
                } else {
                    String err = "Cannot load resources for tomcat service\n";
                    err = err + "You must set TOMCAT_HOME";
                    throw new Exception(err);
                }
            } else {
                // the user want to use tomcat4.0
                String catalinaHome = System.getProperty(CATALINA_HOME);
                if (catalinaHome != null) {
                    File catalinaHomeDir = new File(catalinaHome);
                    jurls.merge(getCatalinaResources(catalinaHomeDir));
                } else {
                    String err = "Cannot load resources for tomcat service\n";
                    err = err + "You must set CATALINA_HOME";
                    throw new Exception(err);                    
                }
            }   
        } 

        // add resources for "web" container service if required.
        if (isLaunchService("web")) {
            boolean webClassesLoaded = false;

            // Get the catalina.home directory and add resources of 
            // Catalina if required.
            String catalinaHome = System.getProperty(CATALINA_HOME);
            if (catalinaHome != null) {
                File catalinaHomeDir = new File(catalinaHome);
                jurls.merge(getCatalinaResources(catalinaHomeDir));
                webClassesLoaded = true;
            }
       
            // Get the jetty.home directory and add resources of 
            // Jetty if required.
            //String jettyHome = System.getProperty(JETTY_HOME);
            //if (jettyHome != null) {
            //    File jettyHomeDir = new File(jettyHome);
            //    jurls.merge(getJettyResources(jettyHomeDir));
            //    webClassesLoaded = true;
            //}

            if (!webClassesLoaded) {
                String err = "Cannot load resources for the web container\n";
                err = err + "You must set CATALINA_HOME";
                //err = err + "You must set CATALINA_HOME or JETTY_HOME";
                throw new Exception(err);
            }
        }


        return jurls.toURLs();
    }


    /**
     * Server main routine (Load all the jars needed to start the JOnAS 
     * commands).
     * @param args the list of the args to give to the bootstrap class.
     */
    public static void main(String args[]) {
        try {
            URL[] urls = getResources();
            JClassLoader jonasLoader = new JClassLoader(urls, null);
            
            // Set our classloader in the current thread JVM.
            Thread.currentThread().setContextClassLoader(jonasLoader);

            // Launch the JOnAS server by using our classloader.
            String classToRun = args[0];
            Class clazz = jonasLoader.loadClass(classToRun);
            Class[] argList = new Class[] { args.getClass() };
            Method meth = clazz.getMethod("main", argList);
            String[] newArgs = new String [args.length - 1];
            System.arraycopy(args, 1, newArgs, 0, newArgs.length);
            Object appli = meth.invoke(null, new Object[]{newArgs});
        } catch (Exception e) {
            // We don't use here the trace system.
            System.err.println("JOnAS error: " + e.getMessage());
            System.err.println(SERVER_NAME + " halting");
            e.printStackTrace();
            System.exit(2);
        }
    }
}
