/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * Initial developer(s): Ludovic BERT & Florent BENOIT
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas_ejb.deployment.lib;

//import java
import java.net.URL;
import java.net.URLClassLoader;
import java.net.MalformedURLException;
import java.io.IOException;
import java.io.File;
import java.util.Vector;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Enumeration;

//import jonas_ejb
import org.objectweb.jonas_ejb.deployment.api.BeanDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDesc;
import org.objectweb.jonas_ejb.deployment.api.DeploymentDescException;
import org.objectweb.jonas_ejb.deployment.api.EjbRefDesc;
import org.objectweb.jonas_ejb.deployment.api.EjbLocalRefDesc;
import org.objectweb.jonas_ejb.deployment.api.SessionDesc;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;

//import monolog
import org.objectweb.jonas.common.Log;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;


/**
 * This class provide a way for managing the EjbDeploymentDesc.
 * Note that there is an intance of the EjbDeploymentDescManager on each JOnAS
 * server.
 * @author Ludovic Bert
 * @author Florent Benoit
 */
public class EjbDeploymentDescManager {

    /**
     * Char separator in an ejbLink
     */
    private static final String EJB_LINK_SEPARATOR = "#";

    /**
     * The unique instance of the EjbDeploymentDescManager.
     */
    private static EjbDeploymentDescManager unique;
    
    /**
     * Associates an URL of an ejb-jar to its ejb deployment descriptor.
     * It's the cache.
     */
    private Hashtable urlJarBindings;

    /**
     * Associates an URL to an ear classloader.
     */
    private Hashtable urlEarCLBindings;

    /**
     * Associates an URL to an ejb classloader.
     */
    private Hashtable urlEjbCLBindings;
    
    /**
     * Associates an URL to an alternate DDesc file.
     */
    private Hashtable urlAltDDBindings = null;

    /**
     * Associates an ear classloader to a list of URLs. It is used to know if
     * the ejb-link is authorised. An ejb-link is authorized only in the same
     * ear.
     */
    private Hashtable earCLEjbLinkJar;

    /**
     * Logger for the deployment desc manager.
     */
    private static Logger logger = null;

    /**
     * Contructs a unique new EjbDeploymentDescManager.
     */
    private EjbDeploymentDescManager() {
        urlJarBindings = new Hashtable();
        urlEarCLBindings = new Hashtable();
        urlEjbCLBindings = new Hashtable();
        earCLEjbLinkJar = new Hashtable();
        urlAltDDBindings = new Hashtable();
        logger = Log.getLogger(Log.JONAS_SERVER_PREFIX);
    }

    /**
     * Get an instance of the EjbDeploymentDescManager.
     * @return the instance of the EjbDeploymentDescManager.
     */
    public static EjbDeploymentDescManager getInstance() {
        if (unique == null) {
            unique = new EjbDeploymentDescManager();
        }
        return unique;
    }
 
    /**
     * Get the specified ejb deployment descriptor and put it in the cache
     * if it is not in.
     * @param url the url where to load xml deployment descriptors.
     * @param loaderForCls classloader used to load bean classes.
     * @param loader the parent classloader (the ear classloader). Null if
     * there in the case of an ejb-jar application.
     * @return DeploymentDesc the ejb deployment descriptor. (This method is
     * used for the ear applications).
     * @throws DeploymentDescException when DeploymentDesc cannot be created 
     * with the given files.
     */
    public synchronized DeploymentDesc getDeploymentDesc(URL url, 
                                            ClassLoader loaderForCls,
                                            ClassLoader loader) 
        throws DeploymentDescException {

        // Check in the ejb-link is allowed : the ejb-link must be done
        // on an ejb-jar defined in the ear application.
        if (loader != null) {
            checkEjbLinkAvailable(loader, url);
        }

        DeploymentDesc dd = (DeploymentDesc) urlJarBindings.get(url.getFile());
        if (dd == null) {
            // dd not in cache => load the deployment descriptor.
            dd = loadDeploymentDesc(url, loaderForCls, loader);
        }
        return dd;
    }

    /**
     * Get the deployment descriptor of the ejb-link of the specified url.
     * @param currentUrl the URL of the component (web or bean) 
     * that do an ejb-link.
     * @param urlToLoad the URL of the bean to get the deployment descriptor.
     * @return the deployment descriptor of the specified bean.
     * @throws DeploymentDescException when DeploymentDesc cannot be created 
     * with the given files.
     */
    private DeploymentDesc getDeploymentDesc(URL currentUrl, URL urlToLoad)
        throws DeploymentDescException {

        DeploymentDesc ejbLinkDD = (DeploymentDesc) urlJarBindings.get(urlToLoad.getFile());
        if (ejbLinkDD == null) {
            // ejbLinkDD not in cache => load the deployment descriptor.
            String url = currentUrl.getFile();
            URLClassLoader earLoader = (URLClassLoader) urlEarCLBindings.get(url);
            if (!url.toLowerCase().endsWith(".jar")) {
                String err = "File '" + url + "' is not a jar file";
                throw new DeploymentDescException(err);
            }

            if (earLoader != null) {
                // In the case of an ear application.      

                // Check in the ejb-link is allowed : the ejb-link must be done
                // on an ejb-jar defined in the ear application.
                checkEjbLinkAvailable(earLoader, urlToLoad);
  
                // get the loader for classes (ejb classloader).
                URLClassLoader loaderForCls = (URLClassLoader) urlEjbCLBindings.get(url);
  
                // get the ear classloader.
                URLClassLoader loader = (URLClassLoader) urlEarCLBindings.get(url);
                ejbLinkDD = loadDeploymentDesc(urlToLoad, loaderForCls, loader);
            } else {
                // In the case of a non ear application.  
                if (!currentUrl.getFile().equals(urlToLoad)) {
                    String err = "In '" + url + "' ejb-link is not allowed outside the ejb-jar if you are not in an ear application.";
                    throw new DeploymentDescException(err);
                }
                // else nothing to do. 
            }     
        }
        return ejbLinkDD;
    }

    /**
     * Load the specified ejb deployment descriptor and 
     * put it in cache.
     * @param url where to load xml deployment descriptors.
     * @param loaderForCls classloader used to load bean classes.
     * @param loader the parent classloader (the ear classloader). Null when 
     * not in the case of an ear application.
     * @return DeploymentDesc the ejb deployment descriptor.
     * @throws DeploymentDescException when DeploymentDesc cannot be created 
     * with the given files.
     */
    private DeploymentDesc loadDeploymentDesc(URL url,
                                              ClassLoader loaderForCls, 
                                              ClassLoader loader) 
            throws DeploymentDescException {

        // Check if the ejb-jar exists.
        if (!new File(url.getFile()).exists()) {
            String err = "Cannot get the deployment descriptor for '" + url.getFile() + "'. The file doesn't exist.";
            throw new DeploymentDescException(err);
        }

        //Check if there is an altenate DD for this URL
        URL altDDUrl = (URL) urlAltDDBindings.get(url);


        // ... and get the instance of the DeploymentDesc.
        //If there is an alternate url for the ejb-jar.xml, call the method with this param.
        DeploymentDesc dd = null;
        if (altDDUrl != null) {
            dd = DeploymentDesc.getInstance(url.getFile(), loaderForCls, altDDUrl.getFile());
        } else {
            dd = DeploymentDesc.getInstance(url.getFile(), loaderForCls);
        }

        // ... and put it in cache.
        if (loader != null) {
            // case of ear application.
            urlEjbCLBindings.put(url.getFile(), loaderForCls);
            urlEarCLBindings.put(url.getFile(), loader);   
        }

        BeanDesc[] bd = dd.getBeanDesc();
        for (int j = 0; j < bd.length; j++) {
            
            // Resolve the ejb-link for ejb-ref
            EjbRefDesc[] ejbRef = bd[j].getEjbRefDesc();
            for (int i = 0; i < ejbRef.length; i++) {
                String jndiName = ejbRef[i].getJndiName();
                String ejbLink = ejbRef[i].getEjbLink();
                String ejbRefType = ejbRef[i].getEjbRefType();
                if (ejbLink != null && jndiName == null) {
                    String ejbName = getJndiName(url, ejbLink, ejbRefType, dd, true);
                    ejbRef[i].setJndiName(ejbName);
                }
            }
            
            // Resolve the ejb-link for ejb-local-ref
            EjbLocalRefDesc[] ejbLocalRef = bd[j].getEjbLocalRefDesc();
            for (int i = 0; i < ejbLocalRef.length; i++) {
                String ejblink = ejbLocalRef[i].getEjbLink();
                if (ejblink == null) {
                    String err = "Ejb-link must be specified for ejb-local-ref " + ejbLocalRef[i].getName();
                    throw new DeploymentDescException(err);
                }        
                String ejbRefType = ejbLocalRef[i].getEjbRefType();
                String ejbName = getJndiName(url, ejblink, ejbRefType, dd, false);
                ejbLocalRef[i].setJndiLocalName(ejbName);
            }
        }

        // ... and put it in cache if ear case.
        if (loader != null) {
            // case of ear application.
            urlJarBindings.put(url.getFile(), dd);
        }
        
        return dd;
    }

    /**
     * Return the JNDI name from the ejbLink string.
     * ejbLink format : filename.jar#beanName in the same Ear File
     *                  beanName in the same ejb-jar file. 
     * @param ejbJar the url of the jar being parsed. This is needed because 
     * ejbLink is relative. With the url and the ejbLink, we can know where
     * the file is locate.
     * @param ejbLink the ejbLink tag of an ejb-ref
     * @param ejbType the type of the referenced ejb in the ejb-ref tag.
     * @param deploymentDesc the deployment descriptor of the parsed deploymentDesc.
     * @param isEjbRef true if the jndi name to resolve is an ejb-ref
     * @return the JNDI name if found, null otherwise
     * @throws DeploymentDescException when it failed
     */
    private String getJndiName(URL ejbJar, String ejbLink, String ejbType,
                               DeploymentDesc deploymentDesc, boolean isEjbRef) 
        throws DeploymentDescException {

        // Extract from the ejb link
        //   - the name of the file
        //   - the name of the bean
        String ejbJarLink = null;
        String beanNameLink = null;

        //Same jar ?
        if (ejbLink.indexOf(EJB_LINK_SEPARATOR) == -1) {
            //Link to a bean inside the same ejb-jar file.
            if (deploymentDesc == null) {
                throw new DeploymentDescException("Deployment desc for file ejb-jar '" + ejbJar.getFile() + "' not found");
            }
            BeanDesc bd = deploymentDesc.getBeanDesc(ejbLink);
            if (bd == null) {
                String err = "Ejb-link " + ejbLink + " not found inside the file " + ejbJar.getFile() + ". The bean doesn't exists.";
                throw new DeploymentDescException(err);
            }
            //Check if the type of the ejb-ref is correct.
            checkType(ejbJar, ejbType, bd);
     
            return isEjbRef ? bd.getJndiName() : bd.getJndiLocalName();
        }

        //Ejb-link not in the same ejb-jar file.
        StringTokenizer st = new StringTokenizer(ejbLink, EJB_LINK_SEPARATOR);

        // We must have only two elements after this step, one for the fileName
        // before the # and the name of the bean after the # char
        if (st.countTokens() != 2 || ejbLink.startsWith("#") || 
            ejbLink.endsWith("#")) {

            String err = "Ejb link " + ejbLink + " has a bad format. Correct format :  filename.jar#beanName.";
            throw new DeploymentDescException(err);
        }

        //Get the token
        ejbJarLink = (String) st.nextToken();
        beanNameLink = (String) st.nextToken();

        //Check if ejbJarLink is a jar or not
        if (!ejbJarLink.endsWith(".jar")) {
            String err = "Ejbjar filename " + ejbJarLink + " from the ejb-link " + ejbLink + " has a bad format. Correct format :  filename.jar";
            throw new DeploymentDescException(err);
        }


        // Now construct the URL from the absolute path from the url ejbJar and
        // the relative path from ejbJarLink
        URL ejbJarLinkUrl = null;
        try {
            ejbJarLinkUrl = new File(new File(ejbJar.getFile()).getParent() + File.separator + ejbJarLink).getCanonicalFile().toURL();
        } catch (MalformedURLException mue) {
            String err = "Error when creating an url for the ejb jar filename. Error :" + mue.getMessage();
            throw new DeploymentDescException(err);
        } catch (IOException ioe) {
            String err = "Error when creating/accessing a file. Error :" + ioe.getMessage();
            throw new DeploymentDescException(err);
        }

        // We've got the url
        // Now, We can ask the Deployment Descriptor of this url
        DeploymentDesc dd = getDeploymentDesc(ejbJar, ejbJarLinkUrl);

        //JndiName resolve.
        BeanDesc bd = dd.getBeanDesc(beanNameLink);

        if (bd == null) {
            String err = "Ejb-link " + ejbLink + " not found inside the file " + ejbJar.getFile() + ". The bean doesn't exists";
            throw new DeploymentDescException(err);
        }

        //Check if the type of the ejb-ref is correct.
        checkType(ejbJar, ejbType, bd);

        return isEjbRef ? bd.getJndiName() : bd.getJndiLocalName();
    }

    /**
     * Check if the type of the ejb-ref is correct.
     * @param ejbJar the URL of the ejb-jar being parsed.
     * @param ejbType the type of the ejb-ref (ejb-ref-type).
     * @param bd the descriptor of the referenced bean.
     * @throws DeploymentDescException if the type is incorrect.
     */
    private void checkType(URL ejbJar, String ejbType, BeanDesc bd) 
        throws DeploymentDescException {
  
        if (bd instanceof SessionDesc) {
            if (!ejbType.equalsIgnoreCase("Session")) {
                String err = "Deployment desc '" + ejbJar.getFile() + "' has an incompatible ejb-ref-type: Required Session but found " + ejbType;
                throw new DeploymentDescException(err);
            }
        } else if (bd instanceof EntityDesc) {
            if (!ejbType.equalsIgnoreCase("Entity")) {
                String err = "Deployment desc '" + ejbJar.getFile() + "' has an incompatible ejb-ref-type: Required Entity but found " + ejbType;
                throw new DeploymentDescException(err);
            }
        } else {
            String err = "Deployment desc '" + ejbJar.getFile() + "' has a bad ejb-ref-type.";
            throw new DeploymentDescException(err);
        }
    }

    /**
     * Check if the ejb-link is available in the case of an ear application.
     * @param earClassLoader the ear classloader of the ear application.
     * @param url the url to check.
     * @throws DeploymentDescException if the ejb-link isn't available.
     */
    private void checkEjbLinkAvailable(ClassLoader earClassLoader,
                                       URL url) 
        throws DeploymentDescException {
        Vector v = (Vector) earCLEjbLinkJar.get(earClassLoader);
        if (v != null) {
            if (!v.contains(url.getFile())) {
                String err = "The ejb-link of '" + url.getFile() + "' must be done on an ejb-jar defined in the ear application (application.xml <module><ejb>###</ejb></module>)";
                throw new DeploymentDescException(err);
            }
        } else {
            String err = "setAvailableEjbLinkJar was badly called.";
            throw new DeploymentDescException(err);
        }
    }

    /**
     * Set for the given ear identified by its earClassLoader the list of the
     * ejb-jar that can be in the ejb-link and the optional Desployment Desc.
     * @param earClassLoader the classloader of the ear application.
     * @param urls the list of the URLs of the ear application that can be in
     * the ejb-link.
     * @param altDDs the list of the URLs of the alternate DDs to use if specified.
     */
    public void setAvailableEjbJarsAndAltDDs(ClassLoader earClassLoader, 
                                             URL[] urls,
                                             URL[] altDDs) {
        Vector v = new Vector();
        for (int i = 0; i < urls.length; i++) {
            v.addElement(urls[i].getFile());
            if (altDDs[i] != null) {
                urlAltDDBindings.put(urls[i].getFile(), altDDs[i]);
            }
        }
        earCLEjbLinkJar.put(earClassLoader, v);
    }

    /**
     * Make a cleanup of the cache of deployment descriptor. This method must 
     * be invoked after the ear deployment by the EAR service.
     * @param earClassLoader the URLClassLoader of the ear application to
     * remove from the cache.
     */
    public void removeCache(ClassLoader earClassLoader) {
        Enumeration enum = earCLEjbLinkJar.keys();
 
        Vector v = (Vector) earCLEjbLinkJar.remove(earClassLoader);
        if (v != null) {
            for (int i = 0; i < v.size(); i++) {
                String url = (String) v.elementAt(i);
                urlJarBindings.remove(url);
                urlEarCLBindings.remove(url);
                urlEjbCLBindings.remove(url);
                urlAltDDBindings.remove(url);
            }
        }

        if (earCLEjbLinkJar.size() != 0 ||
            urlJarBindings.size() != 0 ||
            urlEarCLBindings.size() != 0 ||
            urlAltDDBindings.size() != 0 ||
            urlEjbCLBindings.size() != 0) {
            String buffer = earCLEjbLinkJar.size() + " ";
            buffer += urlJarBindings.size() + " ";
            buffer += urlEarCLBindings.size() + " ";
            buffer += urlEjbCLBindings.size() + " ";
            buffer += urlAltDDBindings.size();

            logger.log(BasicLevel.DEBUG, buffer + " there are some elements in cache");
        }
    }

    /**
     * Get the size of the cache (number of entries in the cache).
     * Used only for debugging.
     * @return the size of the cache (number of entries in the cache).
     */
    public int getCacheSize() {
        return earCLEjbLinkJar.size() + urlJarBindings.size() + 
            urlEarCLBindings.size() + urlEjbCLBindings.size() +
            urlAltDDBindings.size();
    }

    /**
     * Return a string representation of the cache. (Used only for debugging).
     * @return a string representation of the cache.
     */
    public String toString() {
        return earCLEjbLinkJar.size() + " " + urlJarBindings.size() + " " + 
            urlEarCLBindings.size() + " " + urlEjbCLBindings.size() + " " +
            urlAltDDBindings.size();
    }

}
